/* ez-blob-basic.c: Basic encoding

   Copyright (C) 2020 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>

#include "ez-blob.h"
#include "ez-blob-basic.h"

/*
 This implements a basic 'high performance' serialisation based on the
 descriptors.

 Fields are encoded in the same order as the descriptor with no
 labelling.  Therefore the same descriptor must be used on both ends.
 Fields may be at arbitrary alignments.

 primitives
  encoded in machine order at native size

 strings
  4-byte length followed by characters.  No nul termination.  A length
  of ~0 indicates a null string pointer.

** not implemented yet
 arrays
  4-byte length followed by data.  ** currently null pointer arrays with zero length are converted to a zero-length array.

 struct
  4-byte length followed by blob-encoded data.

** not implemented yet
 pointer
  4-byte length followed by blob-encoded data.  A length of ~0
  indicates a null pointer.

*/

size_t ez_basic_size(const ez_blob_desc *desc, const void *p) {
	size_t size = 0;

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		int pshift = (d->bd_type & 3);
		int ptype = d->bd_type & EZ_BLOB_TYPE;
		const void *v = (p + d->bd_offset);

		switch (d->bd_type & EZ_BLOB_STORAGE) {
		case EZ_BLOB_SINGLE:
			if (ptype == EZ_BLOB_STRUCT)
				size += 4 + ez_basic_size(d->bd_table, v);
			else
				size += 1 << pshift;
			break;
		case EZ_BLOB_SINGLEP:
			if (ptype == EZ_BLOB_STRUCT)
				size += ez_basic_size(d->bd_table, *(const void **)v);
			else
				abort();
			break;
		case EZ_BLOB_CSTRING:
			if (ptype == EZ_BLOB_INT8)
				size += (*(const char **)v) ? 4 + strlen(*(const char **)v) : 4;
			else
				abort();
			break;
		default:
			abort();
		}
	}

	return size;
}

int ez_basic_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	const char *b = blob->eb_data;
	size_t size = blob->eb_size;
	const char *be = b + size;

	memset(p, 0, desc->bd_offset);

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		int ptype = d->bd_type & EZ_BLOB_TYPE;
		void *v = (p + d->bd_offset);

		switch (d->bd_type & EZ_BLOB_STORAGE) {
		case EZ_BLOB_SINGLE:
			switch (ptype) {
			case EZ_BLOB_INT8:
				*(uint8_t *)v = *(uint8_t *)b;
				b += 1;
				break;
			case EZ_BLOB_INT16:
				*(uint16_t *)v = *(uint16_t *)b;
				b += 2;
				break;
			case EZ_BLOB_INT32:
			case EZ_BLOB_FLOAT32:
				*(uint32_t *)v = *(uint32_t *)b;
				b += 4;
				break;
			case EZ_BLOB_INT64:
			case EZ_BLOB_FLOAT64:
				*(uint64_t *)v = *(uint64_t *)b;
				b += 8;
				break;
			case EZ_BLOB_STRUCT: {
				const ez_blob data = { .eb_size = *(uint32_t *)b, .eb_data = (void *)(b+4) };

				ez_basic_decode_raw(d->bd_table, &data, v);
				b += 4 + data.eb_size;
				break;
			}
			}
			break;
		case EZ_BLOB_SINGLEP:
			abort();
			break;
		case EZ_BLOB_CSTRING:
			if (ptype == EZ_BLOB_INT8) {
				uint32_t size = *(uint32_t *)b;
				if (size != ~0U) {
					*(char **)v = malloc(size+1);
					memcpy(*(char **)v, b+4, size);
					(*(char **)v)[size] = 0;
					b += 4 + size;
				} else {
					*(char **)v = NULL;
					b += 4;
				}
			} else
				abort();
			break;
		default:
			abort();
		}
	}
	if (b != be)
		goto fail;

	return 0;
 fail:
	ez_blob_free_raw(desc, p);
	return -1;
}

void *ez_basic_decode(const ez_blob_desc *d, const ez_blob *blob) {
	void *p = calloc(d->bd_offset, 1);

	if (ez_basic_decode_raw(d, blob, p) != 0) {
		free(p);
		return NULL;
	}

	return p;
}

int ez_basic_encode_raw(const ez_blob_desc *desc, const void *p, ez_blob *blob) {
	char *b = blob->eb_data;
	size_t size = blob->eb_size;
	char *be = b+size;

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		int ptype = d->bd_type & EZ_BLOB_TYPE;
		const void *v = (p + d->bd_offset);

		switch (d->bd_type & EZ_BLOB_STORAGE) {
		case EZ_BLOB_SINGLE:
			switch (ptype) {
			case EZ_BLOB_INT8:
				*(uint8_t *)b = *(uint8_t *)v;
				b += 1;
				break;
			case EZ_BLOB_INT16:
				*(uint16_t *)b = *(uint16_t *)v;
				b += 2;
				break;
			case EZ_BLOB_INT32:
			case EZ_BLOB_FLOAT32:
				*(uint32_t *)b = *(uint32_t *)v;
				b += 4;
				break;
			case EZ_BLOB_INT64:
			case EZ_BLOB_FLOAT64:
				*(uint64_t *)b = *(uint64_t *)v;
				b += 8;
				break;
			case EZ_BLOB_STRUCT: {
				ez_blob data = { .eb_size = ez_basic_size(d->bd_table, v), .eb_data = b + 4 };

				*(uint32_t *)b = data.eb_size;

				ez_basic_encode_raw(d->bd_table, v, &data);
				b += 4 + data.eb_size;
				break;
			}
			}
			break;
		case EZ_BLOB_SINGLEP:
			abort();
			break;
		case EZ_BLOB_CSTRING:
			if (ptype == EZ_BLOB_INT8) {
				const char *s = *(const char **)v;

				if (s) {
					size_t len = strlen(s);
					*(uint32_t *)b = len;
					memcpy(b+4, s, len);
					b[4+len] = 0;
					b += 4 + len;
				} else {
					*(uint32_t *)b = ~0U;
					b += 4;
				}
			} else
				abort();
			break;
		default:
			abort();
		}
	}
	assert(b == be);
	return 0;
}

int ez_basic_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	blob->eb_size = ez_basic_size(d, p);
	blob->eb_data = malloc(blob->eb_size);

	ez_basic_encode_raw(d, p, blob);

	return 0;
}
