/* ez-blob-compiler.c: Compiler for blob descriptors.

   Copyright (C) 2020 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  This generates encode/decode for a specific structure in a manner compatible
  with the 'basic' encoder.  It is about 2-3x faster than using the table based
  routines.

  This also generates an 'obstack' based decoder interface which
  stores all data in an obstack memory allocator - much faster for the
  decode routine.

  usage: ez-blob-compiler object.o descriptor-name > api.c

  First compile the code that contains the descriptor array, then run this on
  the generated object file to extract the compiled descriptor information.  The
  descriptors must be exported symbols.

 */

#define _GNU_SOURCE

#include <time.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ez-blob.h"
#include "ez-elf.h"

static int doinline = 0;
//static int doobstack = 1;
static ez_elf *elf;

void export_size(const ez_blob_desc *desc, const char *name) {
	size_t size = 0;
	int strings = 0;

	printf("size_t %s_size(const void * __restrict p) {\n", name);

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			size += 1;
			break;
		case EZ_BLOB_INT16:
			size += 2;
			break;
		case EZ_BLOB_INT32:
		case EZ_BLOB_FLOAT32:
			size += 4;
			break;
		case EZ_BLOB_INT64:
		case EZ_BLOB_FLOAT64:
			size += 8;
			break;
		case EZ_BLOB_INT8 | EZ_BLOB_CSTRING:
			strings = 1;
			break;
		default:
			fprintf(stderr, "Unsupported type: %d\n", d->bd_type);
			abort();
		}
	}

	if (strings) {
		printf("\tsize_t size = %zd;\n", size);
		printf("\tconst char *s;\n");
		for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
			const ez_blob_desc *d = &desc[i+1];
			switch (d->bd_type) {
			case EZ_BLOB_INT8:
			case EZ_BLOB_INT16:
			case EZ_BLOB_INT32:
			case EZ_BLOB_INT64:
			case EZ_BLOB_FLOAT32:
			case EZ_BLOB_FLOAT64:
				break;
			case EZ_BLOB_INT8 | EZ_BLOB_CSTRING:
				if (doinline) {
					printf("\ts = *(const char **)(p + %d);\n", d->bd_offset);
					printf("\tsize += s ? strlen(s) + 4 : 4;\n");
				} else {
					printf("\tsize += string_size(*(const char **)(p + %d));\n", d->bd_offset);
				}
				break;
			default:
				fprintf(stderr, "Unsupported type: %d\n", d->bd_type);
				abort();
			}
		}
		printf("\treturn size;\n");
	} else {
		printf("\treturn (size_t)%zd;\n", size);
	}

	printf("}\n");
}

void export_encode(const ez_blob_desc *desc, const char *name) {
	printf("void %s_encode_raw(const void * __restrict p, ez_blob *blob) {\n", name);
	printf("\tchar * __restrict b = blob->eb_data;\n");
	printf("\tconst char *s;\n");

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];

		printf("\t// %d: %s\n", d->bd_id, (char *)ez_elf_read_addr(elf, &d->bd_name));

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			printf("\t*(uint8_t *)b = *(uint8_t *)(p + %d); b += 1;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT16:
			printf("\t*(uint16_t *)b = *(uint16_t *)(p + %d); b += 2;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT32:
		case EZ_BLOB_FLOAT32:
			printf("\t*(uint32_t *)b = *(uint32_t *)(p + %d); b += 4;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT64:
		case EZ_BLOB_FLOAT64:
			printf("\t*(uint64_t *)b = *(uint64_t *)(p + %d); b += 8;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT8 | EZ_BLOB_CSTRING:
			if (doinline) {
				printf("\ts = *(const char **)(p + %d);\n", d->bd_offset);
				printf("\tif (s) {\n"
				       "\t\tsize_t len = strlen(s);\n"
				       "\t\t*(uint32_t *)b = len;\n"
				       "\t\tmemcpy(b+4, s, len);\n"
				       "\t\tb += 4 + len;\n"
				       "\t} else {\n"
				       "\t\t*(uint32_t *)b = ~0U;\n"
				       "\t\tb += 4;\n"
				       "\t}\n");
			} else {
				printf("\tb = string_encode(b, *(const char **)(p + %d));\n", d->bd_offset);
			}
			break;
		default:
			fprintf(stderr, "Unsupported type: %d\n", d->bd_type);
			abort();
		}
	}
	printf("\tif (b - blob->eb_data != blob->eb_size) abort();\n");
	printf("}\n");
}

void export_decode(const ez_blob_desc *desc, const char *name, int doos) {
	if (doos) {
		printf("void *%s_decode_os(const ez_blob * __restrict ez_blob *blob, struct obstack * __restrict os) {\n", name);
		printf("\tvoid *p = obstack_alloc(os, %u);\n", desc->bd_offset);
	} else {
		printf("void %s_decode_raw(const ez_blob * __restrict blob, void * __restrict p) {\n", name);
	}
	printf("\tconst char * __restrict b = blob.eb_data;\n");
	printf("\tsize_t len;\n");
	printf("\tchar **sp;\n");
	if (doos)
		printf("\tmemset(p, 0, %u);\n", desc->bd_offset);

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			printf("\t*(uint8_t *)(p + %d) = *(uint8_t *)b; b += 1;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT16:
			printf("\t*(uint16_t *)(p + %d) = *(uint16_t *)b; b += 2;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT32:
		case EZ_BLOB_FLOAT32:
			printf("\t*(uint32_t *)(p + %d) = *(uint32_t *)b; b += 4;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT64:
		case EZ_BLOB_FLOAT64:
			printf("\t*(uint64_t *)(p + %d) = *(uint64_t *)b; b += 8;\n", d->bd_offset);
			break;
		case EZ_BLOB_INT8 | EZ_BLOB_CSTRING:
			if (doinline) {
				printf("\tsp = (char **)(p + %d);\n"
				       "\tlen = *(uint32_t *)b;\n"
				       "\tif (len != ~0U) {\n", d->bd_offset);
				if (doos)
					printf("\t\t*sp = obstack_copy0(os, b+4, len);\n");
				else
					printf("\t\tchar *s = malloc(len+1);\n"
					       "\t\tmemcpy(s, b+4, len);\n"
					       "\t\ts[len] = 0;\n"
					       "\t\t*sp = s;\n");
				printf("\t\tb += 4 + len;\n"
				       "\t} else {\n"
				       "\t\t*sp = NULL;\n"
				       "\t\tb += 4;\n"
				       "\t}\n");
			} else {
				if (doos)
					printf("\tb = string_decodeos(b, (char **)(p + %d), os);\n", d->bd_offset);
				else
					printf("\tb = string_decode(b, (char **)(p + %d));\n", d->bd_offset);
			}
			break;
		default:
			fprintf(stderr, "Unsupported type: %d\n", d->bd_type);
			abort();
		}
	}
	printf("\tif (b - blob->eb_data != blob->eb_size) abort();\n");
	if (doos)
		printf("\treturn p;\n");
	printf("}\n");
}

int main(int argc, char **argv) {
	int gen = ~0;
	int header = 0;
	char *file = NULL;
	int tables = argc;

	for (int i=1;i<argc;i++) {
		char *cmd = argv[i];

		if (strcmp(cmd, "--generate") == 0 || strcmp(cmd, "-g") == 0) {
			char *genset = argv[i+1];
			int dogen = 0;
			while (genset) {
				char *s = genset;
				char *c = strchr(genset, ',');

				if (c) {
					*c = 0;
					genset = c + 1;
				} else {
					genset = NULL;
				}

				if (strcmp(s, "all") == 0)
					dogen = ~0;
				else if (strcmp(s, "size") == 0)
					dogen |= 1;
				else if (strcmp(s, "encode") ==0)
					dogen |= 2;
				else if (strcmp(s, "decode") == 0)
					dogen |= 4;
				else if (strcmp(s, "decode_obstack") == 0)
					dogen |= 8;
				else if (strcmp(s, "header") == 0)
					header = 1;
			}
			if (dogen)
				gen = dogen;
			i += 1;
		} else {
			file = cmd;
			tables = i + 1;
			break;
		}
	}

	if (!file) {
		fprintf(stderr, "No file supplied");
		return 1;
	}

	if (gen & 8)
		printf("#include <obstack.h>\n");

	if (!doinline && !header) {
		printf("#include <stdint.h>\n");
		printf("#include <libeze/ez-blob.h>\n");
		printf("#include <string.h>\n");
		printf("#include <stdlib.h>\n");

		// should just be another link file i suppose
		printf("static char *string_encode(char * __restrict b, const char *s) {\n"
		       "\tif (s) {\n"
		       "\t\tsize_t len = strlen(s);\n"
		       "\t\t*(uint32_t *)b = len;\n"
		       "\t\tmemcpy(b+4, s, len);\n"
		       "\t\tb += 4 + len;\n"
		       "\t} else {\n"
		       "\t\t*(uint32_t *)b = ~0U;\n"
		       "\t\tb += 4;\n"
		       "\t}\n"
		       "\treturn b;\n"
		       "}\n");
		printf("static const char *string_decode(const char * __restrict b, char **sp) {\n"
		       "\tsize_t len = *(uint32_t *)b;\n"
		       "\tif (len != ~0U) {\n"
		       "\t\tchar *s = malloc(len+1);\n"
		       "\t\tmemcpy(s, b+4, len);\n"
		       "\t\ts[len] = 0;\n"
		       "\t\t*sp = s;\n"
		       "\t\tb += 4 + len;\n"
		       "\t} else {\n"
		       "\t\t*sp = NULL;\n"
		       "\t\tb += 4;\n"
		       "\t}\n"
		       "\treturn b;\n"
		       "}\n");
		if (gen & 8)
			printf("static const char *string_decodeos(const char * __restrict b, char **sp, struct obstack *os) {\n"
			       "\tsize_t len = *(uint32_t *)b;\n"
			       "\tif (len != ~0U) {\n"
			       "\t\t*sp = obstack_copy0(os, b+4, len);\n"
			       "\t\tb += 4 + len;\n"
			       "\t} else {\n"
			       "\t\t*sp = NULL;\n"
			       "\t\tb += 4;\n"
			       "\t}\n"
			       "\treturn b;\n"
			       "}\n");
		printf("static size_t string_size(const char *s) {\n"
		       "\treturn s ? strlen(s) + 4 : 4;\n"
		       "}\n");
	}

	elf = ez_elf_open(file);

	for (int i=tables; i<argc; i++) {

		int dsize;
		const ez_blob_desc *desc;

		ez_elf_addr(elf, argv[i], &dsize, (const void **)&desc);

		//printf("Desc %p, size %d\n", desc, dsize);
		const char *name = ez_elf_read_addr(elf, &desc->bd_name);
		char *xname = strdupa(name);

		if (strncmp(xname, "struct ", 7) == 0)
			xname[6] = '_';

		if (header) {
			if (gen & 1)
				printf("size_t %s_size(const void *p);\n", xname);
			if (gen & 2)
				printf("void %s_encode_raw(const void *p, ez_blob *blob);\n", xname);
			if (gen & 4)
				printf("void %s_decode_raw(const ez_blob *blob, void *p);\n", xname);
			if (gen & 8)
				printf("void *%s_decode_os(const ez_blob *blob, struct obstack *os)\n", xname);
		} else {
			if (gen & 1)
				export_size(desc, xname);
			if (gen & 2)
				export_encode(desc, xname);
			if (gen & 4)
				export_decode(desc, xname, 0);
			if (gen & 8)
				export_decode(desc, xname, 1);
		}

	}
	ez_elf_close(elf);

	return 0;
}
