/* ez-blob-io.c: Common i/o utilities for serialisers

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "ez-blob.h"
#include "ez-blob-io.h"

//#define ABORT() abort()
#define ABORT()

void *blobio_reserve(struct ez_blobio * __restrict io, size_t len) {
	size_t size = io->size;
	size_t to = io->index + len;
	void *v;

	if (io->error)
		return NULL;

	switch (io->mode) {
	case BLOBIO_WRITE_ALLOC:
		if (to > size) {
			void *alloc;

			do {
				size = size ? size * 2 : 256;
			} while (to > size);

			io->size = size;
			alloc = realloc(io->data, size);

			if (!alloc) {
				io->error = ENOMEM;
				printf("realloc failed\n");
				ABORT();
				return NULL;
			}

			io->data = alloc;
		}

		v = io->index + io->data;
		io->index = to;
		return v;

	case BLOBIO_WRITE_FIXED:
		if (to > size) {
			io->error = EOVERFLOW;
			ABORT();
			return NULL;
		}
		v = io->index + io->data;
		io->index = to;
		return v;

	case BLOBIO_WRITE_SIZE:
		io->index = to;
		return NULL;
	default:
		io->error = EINVAL;
		return NULL;
	}
}

void *blobio_take(struct ez_blobio * __restrict io, size_t len) {
	size_t to = io->index + len;

	if (to <= io->size) {
		void *v = io->data + io->index;
		io->index = to;
		return v;
	} else {
		ABORT();
		io->index = io->size;
		io->error = EOVERFLOW;
		return NULL;
	}
}

void blobio_write_align(struct ez_blobio *io, unsigned int step) {
	int skip = (step - io->index) & (step-1);
	uint8_t *v;

	if (skip && (v = blobio_reserve(io, skip)))
		memset(v, 0, skip);
}

void blobio_read_align(struct ez_blobio *io, unsigned int step) {
	int skip = (step - io->index) & (step-1);

	blobio_take(io, skip);
}

#ifndef BLOBIO_INLINE
 void blobio_write(struct ez_blobio *io, const void *data, size_t len) {
	void *v = blobio_reserve(io, len);
	if (v) memcpy(v, data, len);
}

void blobio_writeb(struct ez_blobio *io, uint8_t val) {
	uint8_t *v = blobio_reserve(io, 1);
	if (v) *v = val;
}

void blobio_write8(struct ez_blobio *io, uint8_t val) {
	uint8_t *v = blobio_reserve(io, 1);
	if (v) *v = val;
}

void blobio_write16(struct ez_blobio *io, uint16_t val) {
	uint16_t *v = blobio_reserve(io, 2);
	if (v) *v = val;
}

void blobio_write32(struct ez_blobio *io, uint32_t val) {
	uint32_t *v = blobio_reserve(io, 4);
	if (v) *v = val;
}

void blobio_write64(struct ez_blobio *io, uint64_t val) {
	uint64_t *v = blobio_reserve(io, 8);
	if (v) *v = val;
}

void blobio_writef(struct ez_blobio *io, float val) {
	float *v = blobio_reserve(io, 4);
	if (v) *v = val;
}

void blobio_writed(struct ez_blobio *io, double val) {
	double *v = blobio_reserve(io, 8);
	if (v) *v = val;
}

uint8_t blobio_readi8(struct ez_blobio *io) {
	uint8_t *v = blobio_take(io, 1);
	return v ? *v : 0;
}

uint16_t blobio_readi16(struct ez_blobio *io) {
	uint16_t *v = blobio_take(io, 2);
	return v ? *v : 0;
}

uint32_t blobio_readi32(struct ez_blobio *io) {
	uint32_t *v = blobio_take(io, 4);
	return v ? *v : 0;
}

uint64_t blobio_readi64(struct ez_blobio *io) {
	uint64_t *v = blobio_take(io, 8);
	return v ? *v : 0;
}

float blobio_readf(struct ez_blobio *io) {
	float *v = blobio_take(io, 4);
	return v ? *v : 0.0f;
}

double blobio_readd(struct ez_blobio *io) {
	double *v = blobio_take(io, 8);
	return v ? *v : 0.0;
}
#endif
