/* ez-blob-io.h: Common i/o utilities for serialisers

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

// Grow output as necessary
#define BLOBIO_WRITE_ALLOC 0
// Writing to pre-existing buffer
#define BLOBIO_WRITE_FIXED 1
// Calculating size only
#define BLOBIO_WRITE_SIZE 2
// Read mode
#define BLOBIO_READ 3

struct ez_blobio {
	size_t index;
	uint8_t *data;

	// for read: upper limit
	// for write: allocation size
	size_t size;

	// on read, non-zero means invalid
	int error;
	// io mode.
	int mode;
};

/**
 * Reserve write-space.
 *
 * For BLOBIO_WRITE_ALLOC mode this will fail if any allocation fails.
 *
 * For BLOBIO_WRITE_FIXED this will fail if there is a buffer
 * overflow.
 *
 * For BLOBIO_WRITE_SIZE this will never fail.
 *
 * @param io
 * @param len length required
 * @return base of memory, or NULL on error.
 */
void *blobio_reserve(struct ez_blobio *io, size_t len);

/**
 * Take read-space.
 *
 * For BLOBIO_READ this will fail if there is a buffer underflow.
 *
 * @param io
 * @param len Length required.
 * @return base of memory or NULL on error.
 */
void *blobio_take(struct ez_blobio *io, size_t len);

/**
 * Align write output.
 *
 * Data is padded with 0 bytes.
 *
 * @param io
 * @param step must be power of 2.
 */
void blobio_write_align(struct ez_blobio *io, unsigned int step);

/**
 * Align read input.
 *
 * @param io
 * @param step must be power of 2.
 */
void blobio_read_align(struct ez_blobio *io, unsigned int step);

/**
 * Dump binary data in human-readable hex+ascii format.
 * @todo rename ez_blob_dump(prefix, ez_blob)?
 */
void blobio_dumphex(const char *data, size_t size, const char *prefix);

/*
  The read and write functions read or write the type and silently
  perform no-operation on error.  The error can be checked via
  io.error being non-zero.

  @todo I don't think there's any point in making these inline if it just
  calls an external function anyway.
 */

#define BLOBIO_INLINE
#ifdef BLOBIO_INLINE

static __inline__  void blobio_write(struct ez_blobio *io, const void *data, size_t len) {
	void *v = blobio_reserve(io, len);
	if (v) memcpy(v, data, len);
}

static __inline__ void blobio_writeb(struct ez_blobio *io, uint8_t val) {
	uint8_t *v = blobio_reserve(io, 1);
	if (v) *v = val;
}

static __inline__ void blobio_write8(struct ez_blobio *io, uint8_t val) {
	uint8_t *v = blobio_reserve(io, 1);
	if (v) *v = val;
}

static __inline__ void blobio_write16(struct ez_blobio *io, uint16_t val) {
	uint16_t *v = blobio_reserve(io, 2);
	if (v) *v = val;
}

static __inline__ void blobio_write32(struct ez_blobio *io, uint32_t val) {
	uint32_t *v = blobio_reserve(io, 4);
	if (v) *v = val;
}

static __inline__ void blobio_write64(struct ez_blobio *io, uint64_t val) {
	uint64_t *v = blobio_reserve(io, 8);
	if (v) *v = val;
}

static __inline__ void blobio_writef(struct ez_blobio *io, float val) {
	float *v = blobio_reserve(io, 4);
	if (v) *v = val;
}

static __inline__ void blobio_writed(struct ez_blobio *io, double val) {
	double *v = blobio_reserve(io, 8);
	if (v) *v = val;
}

static __inline__ uint8_t blobio_readi8(struct ez_blobio *io) {
	uint8_t *v = blobio_take(io, 1);
	return v ? *v : 0;
}

static __inline__ uint16_t blobio_readi16(struct ez_blobio *io) {
	uint16_t *v = blobio_take(io, 2);
	return v ? *v : 0;
}

static __inline__ uint32_t blobio_readi32(struct ez_blobio *io) {
	uint32_t *v = blobio_take(io, 4);
	return v ? *v : 0;
}

static __inline__ uint64_t blobio_readi64(struct ez_blobio *io) {
	uint64_t *v = blobio_take(io, 8);
	return v ? *v : 0;
}

static __inline__ float blobio_readf(struct ez_blobio *io) {
	float *v = blobio_take(io, 4);
	return v ? *v : 0.0f;
}

static __inline__ double blobio_readd(struct ez_blobio *io) {
	double *v = blobio_take(io, 8);
	return v ? *v : 0.0;
}

#else

void blobio_write(struct ez_blobio *io, const void *data, size_t len);
void blobio_writeb(struct ez_blobio *io, uint8_t val);
void blobio_write8(struct ez_blobio *io, uint8_t val);
void blobio_write16(struct ez_blobio *io, uint16_t val);
void blobio_write32(struct ez_blobio *io, uint32_t val);
void blobio_write64(struct ez_blobio *io, uint64_t val);
void blobio_writef(struct ez_blobio *io, float val);
void blobio_writed(struct ez_blobio *io, double val);

uint8_t blobio_readi8(struct ez_blobio *io);
uint16_t blobio_readi16(struct ez_blobio *io);
uint32_t blobio_readi32(struct ez_blobio *io);
uint64_t blobio_readi64(struct ez_blobio *io);
float blobio_readf(struct ez_blobio *io);
double blobio_readd(struct ez_blobio *io);

#endif
