/* ez-blob-tagz.c: A tagged blob serialiser.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  Tagged blob encoding.

  This is a streamable self-describing byte-oriented binary format.
  It is a general purpose format and supports a super-set of the
  ez_blob descriptor.  It supports primitive and struct types and
  sequences thereof and there is room for extension.

  Each item beings with a descriptor byte, then followed by a tag id,
  a possible count, and the payload.

 xxxxttcc control byte

  xxxx type code

   0 uint8       unsigned int, value zero-extended
   1 uint16
   2 uint32
   3 uint64
   - reserved
   5 float16
   6 float32
   7 float64
   - reserved
   f struct

  note that for int/float types, code&3 == log2 of element size in bytes

  tt log2 of tag size in bytes

   0 1 byte
   1 2 byte
   2 4 byte
   3 use the count size for tt instead, and there is no count byte

  cc log2 of count size in bytes, used to indicate sequence length or non-sequence.

   0 1 byte
   1 2 byte
   2 4 byte
   3 8 byte

   ff is struct-end code

 A header is a control byte followed by an optional 1/2/4 byte-length tag,
 followed by an optional 1/2/4/8 byte-length count.

 A structure payload is a list of tagged fields until a struct-end
 code.  A structure sequence is a list of count struct-encoded blocks.

 Integers can be stored in the smallest number of bytes, i.e. with
 all leading $00 bytes removed.

 Possible extension is signed types and ignoring leading $ff bytes.

ez_blob requirements:

  All fields must have an id.
  The id for each field must be unique for any given struct.
  The ids must be in incremental order.
  Arrays must be stored at their natural size.

final note:

  This is a bit over-engineered.

*/

/*
  Support matrix:

	       	     single  singlep vector  vectorp list    listp   cstring
                             ?null 1                         ?null   ?null

      int8  	        x      	       	x 2    	       	       	       	x 3
      int16 	        x	 	x 2
      int32 	        x	 	x 2
      int64 	        x	 	x 2
      float 	        x	 	x
      double	        x	 	x
      struct	        x      	x      	       	       	x     	x

  1. All pointers are treated as nullable.  The annotation is ignored.
  2. All bytes of integers are written with no prefix removal.
  3. ??Strings are written containing the trailing 0 byte??

*/

#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "ez-list.h"

#include "ez-blob.h"
#include "ez-blob-tagz.h"
#include "ez-blob-io.h"

//#define D(x) do { x; fflush(stdout); } while (0)
#define D(x)

// These are EZ_* types << 4
#define EZT_INT8    0x00
#define EZT_INT16   0x10
#define EZT_INT32   0x20
#define EZT_INT64   0x30

#define EZT_FLOAT16 0x50
#define EZT_FLOAT32 0x60
#define EZT_FLOAT64 0x70
#define EZT_STRUCT  0xf0

#define EZT_TYPE    0xf0

#define EZT_COUNTSHIFT 0
#define EZT_TAGSHIFT   2
#define EZT_DATASHIFT  4

#define EZT_COUNTSIZE  (0x3 << EZT_COUNTSHIFT)
#define EZT_TAGSIZE    (0x3 << EZT_TAGSHIFT)
#define EZT_DATASIZE   (0x3 << EZT_DATASHIFT)

#define EZT_NOCOUNT (0x3 << EZT_TAGSHIFT)

#define EZT_END    0xff // well, end-struct

static void blobio_writei(struct ez_blobio *io, int sc, uint64_t val) {
	switch (sc) {
	case 0:
		blobio_write8(io, val);
		break;
	case 1:
		blobio_write16(io, val);
		break;
	case 2:
		blobio_write32(io, val);
		break;
	case 3:
		blobio_write64(io, val);
		break;
	}
}

// log2 of significant bytes in value
static int size_code(uint64_t size) {
	if (size <= 0xff)
		return 0x0;
	else if (size <= 0xffff)
		return 0x1;
	else if (size <= 0xffffffff)
		return 0x2;
	else
		return 0x3;
}

static int tag_code(uint32_t size) {
	if (size <= 0xff)
		return 0x0;
	else if (size <= 0xffff)
		return 0x1;
	else
		return 0x2;
}

// tag single item, tt=11 and cc=tag
static void blobio_writet(struct ez_blobio *io, uint8_t type, uint32_t tag) {
	int tc = tag_code(tag);

	blobio_writeb(io, type | EZT_NOCOUNT | (tc << EZT_COUNTSHIFT));
	blobio_writei(io, tc, tag);
}

// tag counted item, tt=tag, cc=count
static void blobio_writetc(struct ez_blobio *io, uint8_t type, uint32_t tag, uint64_t count) {
	int tc = tag_code(tag);
	int cc = size_code(count);

	blobio_writeb(io, type | (tc << EZT_TAGSHIFT) | (cc << EZT_COUNTSHIFT));
	blobio_writei(io, tc, tag);
	blobio_writei(io, cc, count);
}

static void tagz_encode_struct(struct ez_blobio *io, int id, const ez_blob_desc *desc, const void *p) {
	blobio_writet(io, EZT_STRUCT, id);

	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *v = (p + d->bd_offset);
		uint64_t val = 0;
		uint64_t count = 0;
		const void *ptr = NULL;

		switch (d->bd_type & (EZ_BLOB_TYPE | EZ_BLOB_STORAGE)) {
		case EZ_BLOB_SINGLE | EZ_BLOB_INT8:
			val = *(uint8_t *)v;
			goto doint;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT16:
			val = *(uint16_t *)v;
			goto doint;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT32:
			val = *(uint32_t *)v;
			goto doint;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT64:
			val = *(uint64_t *)v;
		doint:
			if (val) {
				int sc = size_code(val);

				blobio_writet(io, EZT_INT8 | (sc << EZT_DATASHIFT), d->bd_id);
				blobio_writei(io, sc, val);
			}
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT32:
			blobio_writet(io, EZT_FLOAT32, d->bd_id);
			blobio_writef(io, *(const float *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT64:
			blobio_writet(io, EZT_FLOAT64, d->bd_id);
			blobio_writed(io, *(const double *)v);
			break;

		case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64:
			count = ((ez_blob *)v)->eb_size;
			if (count) {
				int sc = (d->bd_type & EZ_BLOB_SIZE);
				int tc = (d->bd_type & EZ_BLOB_TYPE) << 4;

				ptr = ((ez_blob *)v)->eb_data;
				blobio_writetc(io, tc, d->bd_id, count);
				blobio_write(io, ptr, count << sc);
			}
			break;

		case EZ_BLOB_CSTRING | EZ_BLOB_INT8:
			// NULL strings are the default, so implement same but allow for empty strings
			v = *(const void **)v;
			if (v) {
				count = strlen(v);
				blobio_writetc(io, EZT_INT8, d->bd_id, count);
				blobio_write(io, v, count);
			}
			break;

		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
			v = *(const void **)v;
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
			if (v)
				tagz_encode_struct(io, d->bd_id, d->bd_table, v);
			break;

		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
			v = *(const void **)v;
		case EZ_BLOB_STRUCT | EZ_BLOB_LIST:
			if (v) {
				count = ez_list_size((ez_list *)v);
				if (count) {
					blobio_writetc(io, EZT_STRUCT, d->bd_id, count);

					for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w); n; w = n, n=ez_node_succ(n)) {
						// which id to use?
						tagz_encode_struct(io, d->bd_table[0].bd_id, d->bd_table, w);
					}
				}
			}
			break;
		default:
			io->error = ENOTSUP;
			return;
		}

	}

	blobio_writeb(io, EZT_END);
}

size_t ez_tagz_size(const ez_blob_desc *d, const void *p) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_SIZE
	};
	tagz_encode_struct(&io, d->bd_id, d, p);

	return io.index;
}

int ez_tagz_encode_raw(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->eb_data,
		.size = blob->eb_size,
		.mode = BLOBIO_WRITE_FIXED
	};

	tagz_encode_struct(&io, d->bd_id, d, p);

	return io.error;
}

int ez_tagz_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_ALLOC
	};

	tagz_encode_struct(&io, d->bd_id, d, p);

	if (!io.error) {
		blob->eb_size = io.index;
		blob->eb_data = io.data;
		return 0;
	}

	free(io.data);
	blob->eb_data = NULL;
	blob->eb_size = 0;

	return io.error;
}

static int blobio_readb(struct ez_blobio *io) {
	uint8_t *v = blobio_take(io, 1);
	return v ? *v : EZT_END;
}

static uint64_t blobio_readi(struct ez_blobio *io, int sc) {
	switch (sc) {
	case 0:
		return blobio_readi8(io);
	case 1:
		return blobio_readi16(io);
	case 2:
		return blobio_readi32(io);
	case 3:
		return blobio_readi64(io);
	}
	return 0;
}

static void tagz_decode_struct(struct ez_blobio *io, const ez_blob_desc *desc, void *p);
static void tagz_decode_fields(struct ez_blobio *io, const ez_blob_desc *desc, void *p);

static unsigned int tagz_blob_type(int h) {
	return (h >> 4) & EZ_BLOB_TYPE;
}

static void tagz_decode_fields(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	uint8_t h;
	const ez_blob_desc *dscan = desc ? desc + 1 : desc;
	const ez_blob_desc *dend = desc ? desc + desc->bd_length + 1 : desc;

	while ((h = blobio_readb(io)) != EZT_END) {
		uint32_t ftag;
		uint64_t fcount;
		int sc = (h >> EZT_DATASHIFT) & 3;
		int tc = (h >> EZT_TAGSHIFT) & 3;
		int cc = (h >> EZT_COUNTSHIFT) & 3;
		int use_count = tc != 3;

		tc = tc != 3 ? tc : cc;
		ftag = blobio_readi(io, tc);
		fcount = use_count ? blobio_readi(io, cc) : 1;

		// This forces the tags to be in order
		// It ensures each field can only be visited once
		const ez_blob_desc *d = NULL;
		{
			while (dscan < dend && dscan->bd_id < ftag)
				dscan++;
			if (dscan < dend && dscan->bd_id == ftag)
				d = dscan++;
		}

		if (p && d) {
			void *v = p + d->bd_offset;

			unsigned int ttype = tagz_blob_type(h);
			unsigned int btype = d->bd_type & EZ_BLOB_TYPE;

			if (use_count) {
				// Type must match exactly
				if (ttype != btype) {
					io->error = EILSEQ;
					return;
				}

				switch (d->bd_type & (EZ_BLOB_STORAGE | EZ_BLOB_TYPE)) {
					/* typed vectors */
				case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
				case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
				case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
				case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
				case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
				case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64: {
					size_t size = fcount << (d->bd_type & EZ_BLOB_SIZE);
					void *src = blobio_take(io, size);

					if (src) {
						void *mem = malloc(size);

						if (mem) {
							memcpy(mem, src, size);
							((ez_blob *)v)->eb_data = mem;
							((ez_blob *)v)->eb_size = fcount;
							D(printf("vector %d [%d]'\n", d->bd_id, fcount));
						} else {
							io->error = ENOMEM;
							return;
						}
					}
					break;
				}

					/* C string */
				case EZ_BLOB_CSTRING | EZ_BLOB_INT8: {
					size_t size = fcount;
					void *src = blobio_take(io, size);
					if (src) {
						char *mem = malloc(size+1);

						if (mem) {
							memcpy(mem, src, size);
							mem[size] = 0;
							*(char **)v = mem;
						} else {
							io->error = ENOMEM;
							return;
						}
					}
					break;
				}

					/* List */
				case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
					*(void **)v = malloc(sizeof(ez_list));
					v = *(void **)v;
					if (!v) {
						io->error = ENOMEM;
						return;
					}
					ez_list_init((ez_list *)v);
				case EZ_BLOB_STRUCT | EZ_BLOB_LIST:
					for (int i=0; !io->error && i < fcount; i++) {
						void *node = ez_blob_alloc(d->bd_table);

						if (node) {
							tagz_decode_struct(io, d->bd_table, node);
							ez_list_addtail((ez_list *)v, node);
						} else {
							io->error = ENOMEM;
							return;
						}
					}
					break;
				default:
					io->error = EILSEQ;
					return;
				}
			} else {
				// Type can either match or the target is a larger or same sized integer
				if (ttype != btype
				    && (ttype > EZ_BLOB_INT64 || btype < ttype)) {
					io->error = EILSEQ;
					return;
				}

				switch (d->bd_type & (EZ_BLOB_STORAGE | EZ_BLOB_TYPE)) {

				case EZ_BLOB_SINGLE | EZ_BLOB_INT8:
					*(uint8_t *)v = blobio_readi(io, sc);
					break;
				case EZ_BLOB_SINGLE | EZ_BLOB_INT16:
					*(uint16_t *)v = blobio_readi(io, sc);
					break;
				case EZ_BLOB_SINGLE | EZ_BLOB_INT32:
					*(uint32_t *)v = blobio_readi(io, sc);
					break;
				case EZ_BLOB_SINGLE | EZ_BLOB_INT64:
					*(uint64_t *)v = blobio_readi(io, sc);
					break;
				case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT32:
					*(float *)v = blobio_readf(io);
					break;
				case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT64:
					*(double *)v = blobio_readd(io);
					break;

				case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
					*(void **)v = ez_blob_alloc(d->bd_table);
					v = *(void **)v;
					if (!v) {
						io->error = ENOMEM;
						return;
					}
				case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
					tagz_decode_fields(io, d->bd_table, v);
					break;

				default:
					io->error = EILSEQ;
					return;
				}
			}
		} else {
			// Just skip fields
			switch (h & EZT_TYPE) {
			case EZT_INT8:
			case EZT_INT16:
			case EZT_INT32:
			case EZT_INT64:
			case EZT_FLOAT32:
			case EZT_FLOAT64:
				if (use_count) {
					blobio_take(io, fcount * (1<<sc));
				} else {
					blobio_take(io, (1<<sc));
				}
				break;
			case EZT_STRUCT:
				if (fcount) {
					for (int i=0; !io->error && i < fcount; i++) {
						tagz_decode_struct(io, NULL, NULL);
					}
				} else {
					tagz_decode_fields(io, NULL, NULL);
				}
				break;
			default:
				io->error = EILSEQ;
				return;
			}
		}
	}
}

static void tagz_decode_struct(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	uint8_t h = blobio_readb(io);
	uint32_t stag;
	int tc = (h >> EZT_TAGSHIFT) & 3;
	int cc = (h >> EZT_COUNTSHIFT) & 3;

	// must be a struct with tc=11 and cc!=11
	if ((h & EZT_TYPE) != EZT_STRUCT
		|| tc != 3 || cc == 3) {
		io->error = EILSEQ;
		return;
	}

	// unused, but could check against desc->bd_id
	stag = blobio_readi(io, cc);
	stag = stag;

	tagz_decode_fields(io, desc, p);
}

int ez_tagz_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	struct ez_blobio io = {
		.size = blob->eb_size,
		.data = blob->eb_data,
		.mode = BLOBIO_READ
	};

	ez_blob_init(desc, p);
	tagz_decode_struct(&io, desc, p);

	if (!io.error)
		return 0;

	ez_blob_free_raw(desc, p);
	return io.error;
}

void *ez_tagz_decode(const ez_blob_desc *desc, const ez_blob *blob) {
	void *p = malloc(desc->bd_offset);

	if (p && ez_tagz_decode_raw(desc, blob, p) != 0) {
		free(p);
		p = NULL;
	}

	return p;
}

/* ********************************************************************** */
// TODO: move to separate .o
// TODO: copy fail behaviour of decode()
static void tagz_dump_struct(struct ez_blobio *io, int depth);
static void tagz_dump_fields(struct ez_blobio *io, int depth);

static void tagz_dump_struct(struct ez_blobio *io, int depth) {
	uint8_t h = blobio_readb(io);
	uint32_t stag;
	int tc = (h >> EZT_TAGSHIFT) & 3;
	int cc = (h >> EZT_COUNTSHIFT) & 3;

	char s[depth+1];
	memset(s, ' ', depth);
	s[depth] = 0;

	// must be a struct with tc=11 and cc!=11
	if ((h & EZT_TYPE) != EZT_STRUCT) {
		printf(" not a struct\n");
		return;
	}
	if (tc != 3) {
		printf(" not a unitary object tc=%d\n", tc);
		return;
	}
	if (cc == 3) {
		printf(" invalid tag size %d\n", cc);
		return;
	}

	stag = blobio_readi(io, cc);

	printf("%s[%02x] %d = {\n", s, h, stag);

	tagz_dump_fields(io, depth+4);
}

static void tagz_dump_fields(struct ez_blobio *io, int depth) {
	int h;

	char s[depth+1];
	memset(s, ' ', depth);
	s[depth] = 0;

	// read fields
	while ((h = blobio_readb(io)) != EZT_END) {
		uint32_t ftag;
		uint64_t fcount;
		int sc = (h >> EZT_DATASHIFT) & 3;
		int tc = (h >> EZT_TAGSHIFT) & 3;
		int cc = (h >> EZT_COUNTSHIFT) & 3;
		int use_count = tc != 3;

		tc = tc != 3 ? tc : cc;
		ftag = blobio_readi(io, tc);
		fcount = use_count ? blobio_readi(io, cc) : 1;

		switch (h & EZT_TYPE) {
		case EZT_INT8:
		case EZT_INT16:
		case EZT_INT32:
		case EZT_INT64:
			if (use_count) {
				void *v = blobio_take(io, fcount * (1<<sc));
				if (v) {
					ez_blob blob = { .eb_size = fcount * (1<<sc), .eb_data = v };
					printf("%s[%02x] integer-%-2d %d [%ld] = {\n", s, h, (1<<sc)*8, ftag, fcount);
					ez_blob_dump(&blob, s);
					printf("%s}\n", s);
				}
			} else {
				printf("%s[%02x] integer-%-2d %d = %02lx\n", s, h, (1<<sc)*8, ftag, blobio_readi(io, sc));
			}
			break;
		case EZT_FLOAT32:
			if (use_count) {
				printf("%sfloat %d [%ld] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++)
					printf("  %f\n", blobio_readf(io));
				printf("%s}\n", s);
			} else {
				printf("%sfloat %d = %f\n", s, ftag, blobio_readf(io));
			}
			break;
		case EZT_FLOAT64:
			if (use_count) {
				printf("%s%d double[%ld] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++)
					printf("%s %f\n", s, blobio_readd(io));
				printf("%s}\n", s);
			} else {
				printf("%s%d double = %f\n", s, ftag, blobio_readd(io));
			}
			break;
		case EZT_STRUCT:
			if (use_count) {
				printf("%sstruct %d [%ld] = {\n", s, ftag, fcount);
				for (int i=0;i<fcount;i++) {
					tagz_dump_struct(io, depth+4);
				}
				printf("%s}\n", s);
			} else {
				tagz_dump_fields(io, depth+4);
			}
			break;
		default:
			printf("%s[%02x] ??\n", s, h);
			break;
		}
	}
	printf("%s[%02x] END $%04lx\n", s, h, io->index-1);
	printf("%s}\n", s);
}

/**
 * ezt format is self-describing so can be dumped directly.
 */
void ez_tagz_dump(ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->eb_data,
		.size = blob->eb_size,
		.mode = BLOBIO_READ
	};

	tagz_dump_struct(&io, 4);
}
