/* ez-blob.h: Serialising description and utils.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <assert.h>

#include "ez-blob.h"
#include "ez-list.h"

void ez_blob_free_raw(const ez_blob_desc *desc, void *p) {
	for (int i=0, dlen = desc->bd_length; i < dlen; i++) {
		const ez_blob_desc *d = &desc[i+1];
		void *v = p + d->bd_offset;
		int st = d->bd_type & EZ_BLOB_STORAGE;
		int dt = d->bd_type & EZ_BLOB_TYPE;
		
		switch (st) {
		case EZ_BLOB_SINGLE:
			if (dt == EZ_BLOB_STRUCT)
				ez_blob_free_raw(d->bd_table, v);
			break;
		case EZ_BLOB_SINGLEP:
			v = *(void **)v;
			if (v) {
				if (dt == EZ_BLOB_STRUCT)
					ez_blob_free_raw(d->bd_table, v);
				free(v);
			}
			break;
		case EZ_BLOB_CSTRING:
			free(((char **)v)[0]);
			break;
		case EZ_BLOB_VECTOR:
			if (dt == EZ_BLOB_STRUCT) {
				// FIXME: free elements
			}
			free(((struct ez_blob *)v)->eb_data);
			break;
		case EZ_BLOB_VECTORP:
			v = *(void **)v;
			if (v) {
				if (dt == EZ_BLOB_STRUCT) {
					// FIXME: free elements
				}
				free(((struct ez_blob *)v)->eb_data);
			}
			break;
		case EZ_BLOB_LIST:
			for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w); n; w = n, n=ez_node_succ(n))
				ez_blob_free(d->bd_table, w);
			break;
		case EZ_BLOB_LISTP:
			v = *(void **)v;
			if (v) {
				for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w); n; w = n, n=ez_node_succ(n))
					ez_blob_free(d->bd_table, w);
				free(v);
			}
			break;
		default:
			break;
		}
	}
}

void ez_blob_free(const ez_blob_desc *d, void *p) {
	if (p) {
		ez_blob_free_raw(d, p);
		free(p);
	}
}

void ez_blob_init(const ez_blob_desc *desc, void *p) {
	memset(p, 0, desc->bd_offset);

	for (int i=0, dlen = desc->bd_length; i < dlen; i++) {
		const ez_blob_desc *d = &desc[i+1];
		void *v = p + d->bd_offset;
		int st = d->bd_type & EZ_BLOB_STORAGE;
		int dt = d->bd_type & EZ_BLOB_TYPE;

		switch (st) {
		case EZ_BLOB_SINGLE:
			if (dt == EZ_BLOB_STRUCT)
				ez_blob_init(d->bd_table, v);
			break;
		case EZ_BLOB_LIST:
			ez_list_init((ez_list *)v);
			break;
		}
	}
}

void *ez_blob_alloc(const ez_blob_desc *d) {
	void *p = malloc(d->bd_offset);
	if (p)
		ez_blob_init(d, p);
	return p;
}
