/* ez-blob.h: Serialising description.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_BLOB_H
#define _EZ_BLOB_H

#include <stddef.h>

/**
 * This is a C structure annotation for implementing a serialiser.
 *
 * It is table driven and can describe most common C data structures,
 * including nested and linked structures.
 *
 * A structure is described by an array of ez_blob_desc.  The first
 * entry must have bd_offset=sizeof(structure), and bd_length=the
 * number of field annotats which follow.
 *
 * As an example:
 *
 * typedef struct dbdisk {
 * 	int id;
 * 	char *uuid;
 * 	char *label;
 * 	char *type;
 * } dbdisk;
 *
 * Has a table such as this, using the init macros.
 *
 * static ez_blob_desc DISK_ENC[] = {
 * 	EZ_BLOB_START(dbdisk, 0, 3),
 * 	EZ_BLOB_STRING(dbdisk, 1, uuid),
 * 	EZ_BLOB_STRING(dbdisk, 2, label),
 * 	EZ_BLOB_STRING(dbdisk, 3, type),
 * };
 *
 * The 'id' field and it's semantics are serialiser dependent.
 *
 */
typedef struct ez_blob_desc {
	unsigned short bd_type;		// type of field
	unsigned short bd_id;		// used to tag encoded entries
	unsigned int bd_offset;		// offset into structure, if type == EZ_BLOB_PK then this is sizeof(struct)
	const char *bd_name;		// name of entry
	union {
		const struct ez_blob_desc *bd_table;// embedded structure or pointer
		unsigned int bd_length;
	};
} ez_blob_desc;

/**
 * Field type
 */
typedef enum ez_blob_desc_type {
	/* Data type */
	EZ_BLOB_INT8,			// (type & 3) = log2(element size)
	EZ_BLOB_INT16,
	EZ_BLOB_INT32,
	EZ_BLOB_INT64,
	EZ_BLOB_FLOAT32 = 6,
	EZ_BLOB_FLOAT64,
	EZ_BLOB_STRUCT = 15,		// bd_table points to another descriptor list

	/* Data Storage */
	EZ_BLOB_SINGLE = 0x00,		// the value at bd_offset
	EZ_BLOB_SINGLEP = 0x10,		// pointer to value at bd_offset
	EZ_BLOB_CSTRING = 0x20,		// 0-terminated string pointer

	EZ_BLOB_VECTOR = 0x40,		// ez_blob at bd_offset
	EZ_BLOB_VECTORP = 0x50,		// pointer to ez_blob at bd_offset
	EZ_BLOB_LIST = 0x60,		// ez_list at bd_offset
	EZ_BLOB_LISTP = 0x70,		// pointer to ez_list at bd_offset

	/* Options */
	EZ_BLOB_ISNULLABLE = 0x100,

	/* meta */
	EZ_BLOB_SIZE = 0x03,		// size mask for int or float types
	EZ_BLOB_TYPE = 0x0f,		// Type mask
	EZ_BLOB_STORAGE = 0xf0,		// Storage mask
	EZ_BLOB_STORAGEB = 4,		// Storage shift

	EZ_BLOB_MAGIC = 0x10f5
} ez_blob_desc_type;

/**
 * This is compatible with lmdb MDB_val
 */
typedef struct ez_blob {
	size_t eb_size;
	union {
		void *eb_data;
		uint8_t eb_data8;
		uint16_t eb_data16;
		uint32_t eb_data32;
		uint64_t eb_data64;
		float *eb_float;
		double *eb_double;
	};
} ez_blob;

#define EZ_BLOB_START(s, id, len) { EZ_BLOB_MAGIC, id, sizeof(s), #s, .bd_length = len }

#define EZ_BLOB_INT8(s, id, f) { EZ_BLOB_INT8 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }
#define EZ_BLOB_INT16(s, id, f) { EZ_BLOB_INT16 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }
#define EZ_BLOB_INT32(s, id, f) { EZ_BLOB_INT32 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }
#define EZ_BLOB_INT64(s, id, f) { EZ_BLOB_INT64 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }
#define EZ_BLOB_FLOAT32(s, id, f) { EZ_BLOB_FLOAT32 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }
#define EZ_BLOB_FLOAT64(s, id, f) { EZ_BLOB_FLOAT64 | EZ_BLOB_SINGLE, id, offsetof(s, f), #f }

#define EZ_BLOB_INT8V(s, id, f) { EZ_BLOB_INT8 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }
#define EZ_BLOB_INT16V(s, id, f) { EZ_BLOB_INT16 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }
#define EZ_BLOB_INT32V(s, id, f) { EZ_BLOB_INT32 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }
#define EZ_BLOB_INT64V(s, id, f) { EZ_BLOB_INT64 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }
#define EZ_BLOB_FLOAT32V(s, id, f) { EZ_BLOB_FLOAT32 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }
#define EZ_BLOB_FLOAT64V(s, id, f) { EZ_BLOB_FLOAT64 | EZ_BLOB_VECTOR, id, offsetof(s, f), #f }

#define EZ_BLOB_STRING(s, id, f) { EZ_BLOB_INT8 | EZ_BLOB_CSTRING, id, offsetof(s, f), #f }
#define EZ_BLOB_STRINGN(s, id, f) { EZ_BLOB_INT8 | EZ_BLOB_CSTRING | EZ_BLOB_ISNULLABLE, id, offsetof(s, f), #f }

#define EZ_BLOB_STRUCT(s, id, f, other) {  EZ_BLOB_STRUCT | EZ_BLOB_SINGLE, id, offsetof(s, f), #f, .bd_table = other }
#define EZ_BLOB_STRUCTP(s, id, f, other) {  EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP, id, offsetof(s, f), #f, .bd_table = other }
#define EZ_BLOB_STRUCTPN(s, id, f, other) {  EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP | EZ_BLOB_ISNULLABLE, id, offsetof(s, f), #f, .bd_table = other }

#define EZ_BLOB_LIST(s, id, f, other) { EZ_BLOB_STRUCT | EZ_BLOB_LIST, id, offsetof(s, f), #f, .bd_table = other }
#define EZ_BLOB_LISTPN(s, id, f, other) { EZ_BLOB_STRUCT | EZ_BLOB_LISTP | EZ_BLOB_ISNULLABLE, id, offsetof(s, f), #f, .bd_table = other }

/**
 * Allocate a new structure to hold a blob.
 *
 * The structure is initialised by ez_blob_init().
 *
 * @param d the descriptor table for the blob.
 */
void *ez_blob_alloc(const ez_blob_desc *d);

/**
 * Initialise a struct to the default state.
 *
 * This sets all values to zero and lists to empty.
 *
 * @param d descriptor table matching struct.
 * @param p pointer to struct base.
 */
void ez_blob_init(const ez_blob_desc *d, void *p);

/**
 * Free struct unmarshalled by ez_blob_decode().
 *
 * @param d descriptor table matching struct.
 * @param p decoded struct from ez_blob_decode().
 */
void ez_blob_free(const ez_blob_desc *d, void *p);

/**
 * Free the contents of a structure, but don't free the container.
 *
 * Use ez_blob_init() if you want to re-use the object.
 *
 * @param p structure to free.  p will not be freed.
 */
void ez_blob_free_raw(const ez_blob_desc *d, void *p);

/**
 * Debug function to dump structured blob contents.
 */
void ez_blob_print(const ez_blob_desc *d, const void *a, int depth);

/**
 * Debug/utility function to dump hexadeximal+ascii display.
 */
void ez_blob_dump(const ez_blob *blob, const char *prefix);

#endif
