
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stddef.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "ez-elf.h"
#include "ez-set.h"

typedef Elf64_Xword elf_addr_t;

struct ez_elf {
	void *base;
	size_t size;
	char *name;

	Elf64_Shdr *symtab; // elf_shdr_name(hdr, ".symtab");

	// Symbol indices
	const char *names;

	// Both tables store the same object.  Only one should free
	// its items.
	struct ez_set byname;
	struct ez_set byaddr;
};

/**
   This same object is stored in both hash tables.

   Use by*_as_entry to map to the object.
 */
struct elf_entry {
	struct ez_node byname;
	struct ez_node byaddr;

	const char *name;
	const Elf64_Sym *sym;
	//elf_add_t addr;
};

static struct elf_entry *byaddr_as_entry(const void *a) {
	return (struct elf_entry *)(a - offsetof(struct elf_entry, byaddr));
}

static struct elf_entry *byname_as_entry(const void *a) {
	return (struct elf_entry *)(a - offsetof(struct elf_entry, byname));
}

static int byaddr_equals(const void *ap, const void *bp) {
	struct elf_entry *a = byaddr_as_entry(ap), *b = byaddr_as_entry(bp);

	return a->sym->st_value == b->sym->st_value;
}

static unsigned int byaddr_hash(const void *ap) {
	const struct elf_entry *a = byaddr_as_entry(ap);

	return ez_hash_int64(a->sym->st_value);
}

static int byname_equals(const void *ap, const void *bp) {
	const struct elf_entry *a = byname_as_entry(ap), *b = byname_as_entry(bp);

	return strcmp(a->name, b->name) == 0;
}

static unsigned int byname_hash(const void *ap) {
	const struct elf_entry *a = byname_as_entry(ap);

	return ez_hash_string(a->name);
}

/**
 * Resolve the starting address of a data block of a section
 */
static void *elf_data(Elf64_Ehdr *hdr, int i) {
	void *base = hdr;
	Elf64_Shdr *shdr = base + hdr->e_shoff + hdr->e_shentsize * i;

	return base + shdr->sh_offset;
}

#if 0
/**
 * Resolve the address of an item in a section table
 */
static void *elf_data_item(Elf64_Ehdr *hdr, Elf64_Shdr *sh, int i) {
	void *base = hdr;

	return base + sh->sh_offset + sh->sh_entsize * i;
}
#endif

/**
 * Resolve section by number
 */
static Elf64_Shdr *elf_shdr(Elf64_Ehdr *hdr, int i) {
	void *base = hdr;

	return base + hdr->e_shoff + hdr->e_shentsize * i;
}

/**
 * Resolve section by name
 */
static Elf64_Shdr *elf_shdr_name(Elf64_Ehdr *hdr, const char *name) {
	const char *names = elf_data(hdr, hdr->e_shstrndx);
	int i;

	for (i=0;i<hdr->e_shnum;i++) {
		Elf64_Shdr *shdr =  elf_shdr(hdr, i);
		int nameid = shdr->sh_name;

		if (nameid == SHN_UNDEF
		    || nameid >= SHN_LORESERVE)
			continue;

		if (!strcmp(name, names + nameid))
			return shdr;
	}

	return NULL;
}

#if 0
/**
 * Resolve programme header by id
 */
static Elf64_Phdr *elf_phdr(Elf64_Ehdr *hdr, int i) {
	void *base = hdr;

	return base + hdr->e_phoff + hdr->e_phentsize * i;
}
#endif

/**
 * Find a reloc hunk that references section
 */
static Elf64_Shdr *elf_reloc_section(Elf64_Ehdr *hdr, int section) {
	int i;

	for (i=0;i<hdr->e_shnum;i++) {
		Elf64_Shdr *shdr = elf_shdr(hdr, i);

		if (shdr->sh_type == SHT_RELA  && shdr->sh_info == section)
			return shdr;
	}

	return NULL;
}

static Elf64_Sym *elf_sym(struct ez_elf *elf, int i) {
	return elf->base + elf->symtab->sh_offset + elf->symtab->sh_entsize * i;
}

static void elf_init(struct ez_elf *elf) {
	Elf64_Ehdr *hdr = elf->base;
	Elf64_Shdr *shdr = elf_shdr_name(hdr, ".symtab");
	void *ss, *se;

	ez_set_init(&elf->byname, byname_hash, byname_equals, free);
	ez_set_init(&elf->byaddr, byaddr_hash, byaddr_equals, NULL);

	if (!shdr) {
		elf->names = NULL;
		return;
	}

	elf->names = elf_data(hdr, shdr->sh_link);

	elf->symtab = shdr;

	ss = elf->base + shdr->sh_offset;
	se = ss + shdr->sh_size;

	for (; ss<se; ss += shdr->sh_entsize) {
		Elf64_Sym *sym = ss;

		if (sym->st_name) {
			struct elf_entry *e;

			e = malloc(sizeof(*e));
			e->name = elf->names + sym->st_name;
			e->sym = sym;

			//printf("entry: %s\n", e->name);

			if (ez_set_get(&elf->byname, &e->byname)) {
				fprintf(stderr, "duplicate symbol name ignored: %s\n", e->name);
				free(e);
				continue;
			}
			ez_set_put(&elf->byname, &e->byname);

			if (sym->st_value) {
				if (ez_set_get(&elf->byaddr, &e->byaddr)) {
					fprintf(stderr, "duplicate symbol addr ignored: %s\n", e->name);
				} else
					ez_set_put(&elf->byaddr, &e->byaddr);
			}
		}
	}
}

/**
 * Open an elf file
 */
struct ez_elf *ez_elf_open(const char *name) {
	int fd;
	void *base;
	struct stat st;
	struct ez_elf *elf = NULL;

	fd = open(name, O_RDONLY);
	if (fd == -1)
		return NULL;

	if (fstat(fd, &st) == -1)
		goto fail;

	base = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (!base)
		goto fail;

	fprintf(stderr, "elf_open('%s') mmap=%p\n", name, base);

	elf = malloc(sizeof(*elf));
	elf->name = strdup(name);
	elf->base = base;
	elf->size = st.st_size;

	elf_init(elf);

 fail:
	close(fd);

	return elf;
}

void ez_elf_close(struct ez_elf *elf) {
	munmap(elf->base, elf->size);

	ez_set_clear(&elf->byaddr);
	ez_set_clear(&elf->byname);

	free(elf);
}

const void *ez_elf_read_addr(struct ez_elf *elf, const void *addr) {
	// find section it belongs to
	Elf64_Ehdr *hdr = elf->base;
	off_t offset = addr - elf->base;
	//const char *names = elf_data(hdr, hdr->e_shstrndx);

	for (int i=0;i<hdr->e_shnum;i++) {
		Elf64_Shdr *shdr = elf_shdr(hdr, i);

		//printf("offset %zd section %d '%s' %zd-%zd\n",
		//      offset, i, names + shdr->sh_name,
		//      shdr->sh_offset, shdr->sh_offset + shdr->sh_size);

		if (offset >= shdr->sh_offset && offset < shdr->sh_offset + shdr->sh_size) {
			offset -= shdr->sh_offset;

			//printf("Section is %d  offset %p\n", i, offset);

			Elf64_Shdr *rhdr = elf_reloc_section(elf->base, i);
			if (rhdr) {
				void *ss, *se;

				ss = elf->base + rhdr->sh_offset;
				se = ss + rhdr->sh_size;
				for (; ss<se; ss += rhdr->sh_entsize) {
					Elf64_Rela *rel = ss;
					int symi = ELF64_R_SYM(rel->r_info);
					Elf64_Sym *sym = elf_sym(elf, symi);

					//printf("check rel offset %p\n", rel->r_offset);

					if (rel->r_offset == offset) {
						//printf(" reloc %08x [%08x] + %08x\n", rel->r_info, rel->r_offset, rel->r_addend);
						switch (ELF64_R_TYPE(rel->r_info)) {
						case R_X86_64_64:
#if 0
							printf("  R_X86_64_64  sym %d '%s'\n", symi, elf->names + sym->st_name);
							printf(" name %08x\n", sym->st_name);
							printf(" valu %08x\n", sym->st_value);
							printf(" size %d\n", sym->st_size);
							printf(" info %02x\n", sym->st_info);
							printf(" sidx %02x\n", sym->st_shndx);
							printf("value: %s\n", elf_data(elf->base, sym->st_shndx) + rel->r_addend);
#endif
							return elf_data(elf->base, sym->st_shndx) + rel->r_addend;
						default:
							printf("unknown reloc type\n");
						}
					}
				}
			}
			break;
		}

		//if (shdr->sh_type == SHT_RELA  && shdr->sh_info == section)
		//	return shdr;
	}

	return NULL;

//	Elf64_Shdr *shdr = base + hdr->e_shoff + hdr->e_shentsize * i;

//	return base + shdr->sh_offset;
}

/**
 * Resolve a raw symbol value, optionally the mapped address and data size
 *
 * TODO: reloc entries?
 */
elf_addr_t ez_elf_addr(struct ez_elf *elf, const char *name, int *sizep, const void **ptr) {
	struct elf_entry key = { .name = name };
	struct ez_node *n;

	n = ez_set_get(&elf->byname, &key.byname);
	if (n) {
		struct elf_entry *e = byname_as_entry(n);
		void *mem = elf_data(elf->base, e->sym->st_shndx) + e->sym->st_value;
		int size = e->sym->st_size;

		if (sizep)
			*sizep = size;
		if (ptr)
			*ptr = mem;

		//ez_elf_reloc(elf, mem + 8);

		/* TODO: find a way to get relocs to actual data
		   Perhaps resolve_addr(), but it's dependent on ?
		   addr -> find section
		   section -> find relocs
		*/

#if 0
		printf(" name %08x\n", e->sym->st_name);
		printf(" valu %08x\n", e->sym->st_value);
		printf(" size %d\n", e->sym->st_size);
		printf(" info %02x\n", e->sym->st_info);
		printf(" sidx %02x\n", e->sym->st_shndx);

		Elf64_Shdr *rhdr = elf_reloc_section(elf->base, e->sym->st_shndx);
		if (rhdr) {
			void *ss, *se;

			ss = elf->base + rhdr->sh_offset;
			se = ss + rhdr->sh_size;
			for (; ss<se; ss += rhdr->sh_entsize) {
				Elf64_Rela *rel = ss;
				int symi = ELF64_R_SYM(rel->r_info);
				Elf64_Sym *sym = elf_sym(elf, symi);

				printf("\n*Check sym range %d-%d value %d\n",
				       e->sym->st_value,
				       e->sym->st_value + e->sym->st_size,
				       rel->r_offset);

				if (rel->r_offset >= e->sym->st_value
				    && rel->r_offset < e->sym->st_value + e->sym->st_size) {

					printf(" reloc %08x [%08x] + %08x\n", rel->r_info, rel->r_offset, rel->r_addend);
					switch (ELF64_R_TYPE(rel->r_info)) {
					case R_X86_64_64:
						printf("  R_X86_64_64  sym %d '%s'\n", symi, elf->names + sym->st_name);
						printf(" name %08x\n", sym->st_name);
						printf(" valu %08x\n", sym->st_value);
						printf(" size %d\n", sym->st_size);
						printf(" info %02x\n", sym->st_info);
						printf(" sidx %02x\n", sym->st_shndx);
						printf("value: %s\n", elf_data(elf->base, sym->st_shndx) + rel->r_addend);
						break;
					default:
						printf("unknown reloc type\n");
					}
				}
			}
		}
#endif
#if 0
		printf("find addr %s\n", name);
#endif
		return e->sym->st_value;
	}

	return 0;
}

/**
 * Resolve a symbol address.
 */
const char *ez_elf_name(struct ez_elf *elf, elf_addr_t addr, int *size) {
	const Elf64_Sym sym = { .st_value = addr };
	struct elf_entry key = { .sym = &sym };
	struct ez_node *n;

	n = ez_set_get(&elf->byaddr, &key.byaddr);
	if (n) {
		struct elf_entry *e = byaddr_as_entry(n);

		if (size)
			*size =e->sym->st_size;
		return e->name;
	}

	return NULL;
}
