/* ez-list.h: Double-linked list implementation (aka 'exec lists')

   Copyright (C) 2010,2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef _EZ_LIST_H
#define _EZ_LIST_H

#include <ez-node.h>

/**
 * ez-list is a low-level double-linked list implementation that works
 * without memory allocation.  
 *
 * o Adding to or removing from either end is an O(1) operation.
 * o Removing any given node is an O(1) operation.
 * o Nodes can only belong to a single list at any given time.
 * o No operations can fail.
 * o No pointers are checked.
 * o The list is not thread-safe.
 *
 * All list functions are implemented as inline functions.
 *
 * This is how to iterate all elements in forward order:

        struct list_item {
		struct ez_node ln;
		// item data
	}

	struct list list = EZ_INIT_LIST(list);

	struct list_item *work = ez_list_head(list);
	struct list_item *next = ez_list_succ(list);

	while (next) {

		// operate on 'work'.
		// ez_node_remove() may be called here

		succ = next;
		next = ez_list_succ(list);
	}

 * For reverse order simply use ez_list_tail() and ez_node_pred().
 *
 * Functions use untyped pointers for nodes for convenience but they
 * must start with struct ez_node.
 *
 * Historical note: this is a re-implementation of Amiga's kernel
 * minimal lists.  A minor change is that functions take and return
 * untyped nodes.
 */

typedef struct ez_list ez_list;


/**
 * A list header.
 *
 * This represents two overlapping nodes which share a common pointer.
 * It must be initialised using ez_list_init() or EZ_INIT_LIST()
 * before use.
 *
 * tail is always NULL.
 *
 * The pointers are volatile to avoid aliasing issues with C99+.
 */
struct ez_list {
	struct ez_node * volatile head;
	struct ez_node * volatile tail;
	struct ez_node * volatile tail_pred;
};

/**
 * Initialise a list statically.
 *
 * struct ez_list list = EZ_INIT_LIST(list);
 */
#define EZ_INIT_LIST(l) { (struct ez_node *)&(l).tail, (void *)0, (struct ez_node *)&l }

/**
 * Initialise a list at runtime.
 * 
 */
static __inline__ void ez_list_init(struct ez_list *l) {
	l->head = (struct ez_node *)&l->tail;
	l->tail = (struct ez_node *)0L;
	l->tail_pred = (struct ez_node *)&l->head;
}

/**
 * Add a node to the head of the list.
 */
static __inline__ void ez_list_addhead(struct ez_list *l, void *np) {
	struct ez_node *n = np, *head = l->head;

	n->u.list.succ = head;
	n->u.list.pred = (struct ez_node *)&l->head;
	head->u.list.pred = n;
	l->head = n;
}

/**
 * Add a node to the tail of the list.
 */
static __inline__ void ez_list_addtail(struct ez_list *l, void *np) {
	struct ez_node *n = np, *tail_pred = l->tail_pred;

	n->u.list.pred = tail_pred;
	n->u.list.succ = (struct ez_node *)&l->tail;
	tail_pred->u.list.succ = n;
	l->tail_pred = n;
}

/**
 * Remove a node from a list.
 *
 * The node must already belong to a list.
 *
 * This performs a software memory barrier after
 * write do workaround aliasing issues with gcc.
 */
static __inline__ void ez_node_remove(void *np) {
	struct ez_node *n = np, *pred = n->u.list.pred, *succ = n->u.list.succ;
	
	pred->u.list.succ = succ;
	succ->u.list.pred = pred;
}

/**
 * Remove a node from the head of the list.
 *
 * @return the head node, or NULL if the list is empty.
 */
static __inline__ void *ez_list_remhead(struct ez_list *l) {
	struct ez_node *n = l->head;
	struct ez_node *s = n->u.list.succ;

	if (s) {
		s->u.list.pred = (struct ez_node *)&l->head;
		l->head = s;
		return n;
	} else
		return 0L;
}

/**
 * Remove a node from the tail of the list.
 *
 * @return the tail node, or NULL if the list is empty.
 */
static __inline__ void *ez_list_remtail(struct ez_list *l) {
	struct ez_node *n = l->tail_pred;
	struct ez_node *p = n->u.list.pred;

	if (p) {
		p->u.list.succ = (struct ez_node *)&l->tail;
		l->tail_pred = p;
		return n;
	} else
		return 0L;
}

/**
 * Insert a node after a given node.
 *
 * @param l list.
 * @param n node to insert.
 * @param p node that is already present in list, or NULL (insert after head).
 */
static __inline__ void ez_list_insert_after(struct ez_list *l, void *n, void *p) {
	struct ez_node *nn = n;
	struct ez_node *pp = p ? p : (struct ez_node *)&l->head;
	struct ez_node *succ = pp->u.list.succ;

	nn->u.list.succ = succ;
	nn->u.list.pred = pp;
	succ->u.list.pred = nn;
	pp->u.list.succ = nn;
}

/**
 * Insert a node before a given node.
 *
 * @param l list.
 * @param n node to insert.
 * @param p node that is already present in list, or NULL (insert before tail).
 */
static __inline__ void ez_list_insert(struct ez_list *l, void *n, void *p) {
	struct ez_node *nn = n;
	struct ez_node *pp = p ? p : (struct ez_node *)&l->tail;
	struct ez_node *pred = pp->u.list.pred;
	
	nn->u.list.succ = pp;
	nn->u.list.pred = pred;
	pred->u.list.succ = nn;
	pp->u.list.pred = nn;
}

/**
 * Find the length of the list.
 *
 * This is an O(n) operation, use ez_list_empty() to check for empty
 * lists.
 */
static __inline__ int ez_list_size(struct ez_list *l) {
	struct ez_node *w, *n;
	int size = 0;

	for (w = l->head, n = w->u.list.succ; n; w = n, n = n->u.list.succ)
		size += 1;

	return size;
}

/**
 * Check if the list is empty.
 *
 * @param returns true (1) if the list is empty.
 */
static __inline__ int ez_list_empty(struct ez_list *l) {
	return l->tail_pred == (struct ez_node *)l;
}

/**
 * Retrieve the successor to the head node (the first user node).
 */
static __inline__ void *ez_list_head(struct ez_list *l) {
	return (void *)l->head;
}

/**
 * Retrieve the predecessor of the tail node (the last user node).
 */
static __inline__ void *ez_list_tail(struct ez_list *l) {
	return (void *)l->tail_pred;
}

/* This is important for -Os. */
static void ez_list_init(struct ez_list *l) __attribute__ ((always_inline));
static void ez_list_addhead(struct ez_list *l, void *np) __attribute__ ((always_inline));
static void ez_list_addtail(struct ez_list *l, void *np) __attribute__ ((always_inline));
static void ez_node_remove(void *np) __attribute__ ((always_inline));
static void *ez_list_remhead(struct ez_list *l) __attribute__ ((always_inline));
static void *ez_list_remtail(struct ez_list *l) __attribute__ ((always_inline));
static void ez_list_insert_after(struct ez_list *l, void *n, void *p) __attribute__ ((always_inline));
static void ez_list_insert(struct ez_list *l, void *n, void *p) __attribute__ ((always_inline));
static int ez_list_empty(struct ez_list *l) __attribute__ ((always_inline));
static void *ez_list_head(struct ez_list *l) __attribute__ ((always_inline));
static void *ez_list_tail(struct ez_list *l) __attribute__ ((always_inline));
 
#endif
