/* ez-tree.h: Balanced (avl) binary tree.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#ifndef _EZ_TREE_H
#define _EZ_TREE_H

#include <ez-node.h>

typedef struct ez_tree ez_tree;

/**
 * Tree structure.
 *
 * root is a dummy node used to avoid special cases in the code.  Use
 * ez_tree_root() to obtain the actual tree root node.
 */
struct ez_tree {
	ez_node root;
	ez_cmp_fn node_cmp;
};

/**
 * Static initialiser for a tree.
 *
 *   struct ez_tree tree = EZ_INIT_TREE(tree, cmp);
 */
#define EZ_INIT_TREE(t, cmp) { .root.u.tree.link[0] = 0, .root.u.tree.link[1] = 0, .node_cmp = cmp }

typedef struct ez_tree_scan ez_tree_scan;

/**
 * Structure used to track search decisions when navigating the tree.
 */
struct ez_tree_scan_info {
	struct ez_node *node;
	int link;
};

/**
 * Maximum depth of tree.
 */
#define EZ_TREE_MAX_DEPTH 33

/**
 * Holds iterator state.  Never holds any allocated resources.
 *
 * Public so it can be allocated on the stack.
 */
struct ez_tree_scan {
	ez_tree *tree;
	int level;
	int scan_dir;
	struct ez_tree_scan_info stack[EZ_TREE_MAX_DEPTH];
};

/**
 * Initialise an empty tree.
 *
 * All tree entries must include an embedded ez_node as the first
 * member.
 *
 * @param tree tree to initialise.
 * @param cmp node comparison function for this tree.
 */
void ez_tree_init(struct ez_tree *tree, ez_cmp_fn cmp);

/**
 * Number of nodes in tree.
 */
static __inline__ int ez_tree_size(ez_tree *tree) {
	// well if it's good enough for Knuth.
	return tree->root.u.tree.link[0];
}

/**
 * Retrieve the root node of the tree.
 */
static __inline__ void *ez_tree_root(ez_tree *tree) {
	return (void *)tree->root.u.tree.link[1];
}

/**
 * Determine if the tree is empty.
 */
static __inline__ int ez_tree_empty(ez_tree *tree) {
	return !tree->root.u.tree.link[1];
}

/**
 * Remove all nodes from a tree.
 *
 * This will be substantially quicker than removing all nodes
 * individually as there is no need to maintain the tree balance
 * incrementally.
 *
 * Any nodes are freed using the node_free function.
 *
 * @param tree
 * @param node_free free function.  Must be supplied.
 */
void ez_tree_clear(ez_tree *tree, ez_free_fn node_free);

/**
 * Retrieve an item by key.
 *
 * @param tree
 * @param key_node is of the same type as the entires in the tree but
 * only needs to contain enough valid fields for node_cmp to function.
 * @return an exact match for the key or NULL if no such key exists.
 */
void *ez_tree_get(ez_tree *tree, const void *key_node);

/**
 * Put a new item into the tree.  This will replace any existing
 * item which matches.
 *
 * @param tree
 * @param node a data node.  The type must start with an ez_node.
 * @return the previous node which matches the same key.
 */
void *ez_tree_put(ez_tree *tree, void *node);

/**
 * Remove an item from the tree by key.
 *
 * @param tree
 * @param node_cmp node compare function.
 * @param key_node node who's quality indicates a match.
 * @return a node matching key_node, or NULL if no such node exists.
 */
void *ez_tree_remove(ez_tree *tree, const void *key_node);

/**
 * Initialise tree scanner for walking the tree in order.
 *
 * This initialises a general purpose iterator.  Typical operations of
 * FIRST, LAST, greater-or-equal, or less-or-equal are supported by
 * choosing appropriate parameters.
 *
 * If a key is supplied in @param key_node then it is used as the
 * reference point for the initialisation.  EZ_LINK_LEFT will find the
 * nearest node less than or equal to the key, and EZ_LINK_RIGHT will
 * find the nearest node greater than or equal to the key.
 *
 * If no key is supplied then EZ_LINK_LEFT finds the left-most node in
 * the whole tree (first), and EZ_LINK_RIGHT finds the right-most node
 * (last).
 *
 * Note that the tree must not be modified in any way whilst the scanner
 * is being used outside of the scanner functions.
 *
 * @param tree initialised tree.
 * @param scan (un)initialised scan state.  This function will initialise it.
 * @param scan_dir Direction of scan.
 * @param key_dir Direction to initialise in.
 * @param key_node optional key node for starting point.
 * @return The found node if any is returned.
 */
void *ez_tree_scan_init_key(ez_tree *tree, ez_tree_scan *scan, enum ez_node_link_t scan_dir, enum ez_node_link_t key_dir, void *key_node);


/**
 * Initialise tree scanner for walking tree in order from an end.
 *
 * This is equivalent of:
 * ez_tree_scan_init_key(tree, scan, dir, 1-dir, NULL);
 * @param tree initialised tree.
 * @param scan (un)initialised scan state.  This function will initialise it.
 * @param scan_dir Direction of scan.
 */
void *ez_tree_scan_init(ez_tree *tree, ez_tree_scan *scan, enum ez_node_link_t scan_dir);

/**
 * Move to the next node in the tree in scan order.
 *
 * @param scan initialsied scan state.
 * @return the next node, or NULL if the boundaries have been reached.
 */
void *ez_tree_scan_next(ez_tree_scan *scan);

/**
 * Move to the previous node in the tree in scan order.
 *
 * @param scan initialsied scan state.
 * @return the next node, or NULL if the boundaries have been reached.
 */
void *ez_tree_scan_prev(ez_tree_scan *scan);

/**
 * Remove the current node.
 *
 * This removes the node and then moves to the next node based on the
 * direction.  If a value of -1 is used for @param dir then no
 * movement will take place the the scan must be re-initialised for
 * any further use.
 *
 * Nodes must not be freed until after they have been removed.
 *
 * @param dir The direction to move in next, or -1 to finish the scan.
 * @return the next node in the specified direction or NULL if none exist.
 */
void *ez_tree_scan_remove(ez_tree_scan *scan, enum ez_node_link_t dir);

/**
 * Replace the current entry.
 *
 * Note that the node inserted must not break the ordering of the
 * tree.
 *
 * @param node New entry.
 * @return The previous entry at this location.
 */
void *ez_tree_scan_put(ez_tree_scan *scan, void *node);

static int ez_tree_empty(ez_tree *tree) __attribute__((always_inline));
static int ez_tree_size(struct ez_tree *tree) __attribute__((always_inline));
static void *ez_tree_root(struct ez_tree *tree) __attribute__((always_inline));


#endif
