
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <assert.h>
#include <errno.h>

#include "ez-blob.h"
#include "ez-list.h"

#include "ez-blob-tagz.h"
#include "ez-blob-xdrn.h"
#include "ez-blob-basic.h"

#include "ez-blob-io.h"

/*
  TODO: some bad data tests.

  This depends a lot on the structure and the format.

  * lengths outside of data
  * counts that will exhaust memory
  * field id's out of range
  * data corruption (wrong output)
  * meta-data corruption (wrong/invalid structure)
  * truncated data
  * too much data

  Primarily it shouldn't crash or corrupt memory.
  With explicit knowledge of format, it should fail on corrupt meta-data.
 */

static int doprint = 1;
static int dolarge = 0;

struct simple {
	char b;
	short s;
	int a;
	int64_t l;
	float c;
	double d;
	char *string;
};

const ez_blob_desc simple_DESC[] = {
	EZ_BLOB_START(struct simple, 0, 7),
	EZ_BLOB_INT8(struct simple, 1, b),
	EZ_BLOB_INT16(struct simple, 2, s),
	EZ_BLOB_INT32(struct simple, 3, a),
	EZ_BLOB_INT64(struct simple, 4, l),
	EZ_BLOB_FLOAT32(struct simple, 5, c),
	EZ_BLOB_FLOAT64(struct simple, 6, d),
	EZ_BLOB_STRING(struct simple, 7, string),
};

struct arrays {
	uint32_t count;
	char *string;
	ez_blob array;
};

static const ez_blob_desc array_DESC[] = {
	EZ_BLOB_START(struct arrays, 0, 3),
	EZ_BLOB_INT32(struct arrays, 1, count),
	EZ_BLOB_STRING(struct arrays, 2, string),
	EZ_BLOB_INT8V(struct arrays, 3, array),
};

struct embedded {
	struct simple s;
	struct arrays a;
};

static const ez_blob_desc embed_DESC[] = {
	EZ_BLOB_START(struct embedded, 0, 2),
	EZ_BLOB_STRUCT(struct embedded, 1, s, simple_DESC),
	EZ_BLOB_STRUCT(struct embedded, 2, a, array_DESC),
};

struct tree {
	struct tree *left;
	struct tree *right;

	char *s;
};

static const ez_blob_desc tree_DESC[] = {
	EZ_BLOB_START(struct tree, 0, 3),
	EZ_BLOB_STRING(struct tree, 1, s),
	EZ_BLOB_STRUCTPN(struct tree, 2, left, tree_DESC),
	EZ_BLOB_STRUCTPN(struct tree, 3, right, tree_DESC),
};

struct listnode {
	ez_node ln;
	char *name;
	ez_blob value;
};

static const ez_blob_desc listnode_DESC[] = {
	EZ_BLOB_START(struct listnode, 0, 2),
	EZ_BLOB_STRING(struct listnode, 1, name),
	EZ_BLOB_INT8V(struct listnode, 2, value),
};

struct list {
	char *name;
	ez_list list;
};

static const ez_blob_desc list_DESC[] = {
	EZ_BLOB_START(struct list, 0, 2),
	EZ_BLOB_STRING(struct list, 1, name),
	EZ_BLOB_LIST(struct list, 2, list, listnode_DESC),
};

struct intsize {
	int16_t value16;
	int32_t value32;
	int64_t value64;
};

static const ez_blob_desc intsize_DESC[] = {
	EZ_BLOB_START(struct intsize, 7, 3),
	EZ_BLOB_INT16(struct intsize, 1, value16),
	EZ_BLOB_INT32(struct intsize, 2, value32),
	EZ_BLOB_INT64(struct intsize, 3, value64),
};

struct stringsize {
	char *value;
};

static const ez_blob_desc stringsize_DESC[] = {
	EZ_BLOB_START(struct stringsize, 8, 1),
	EZ_BLOB_STRING(struct stringsize, 1, value),
};

static void dumphex(const char *data, size_t size, const char *prefix) {
	ez_blob blob = { .eb_size = size, .eb_data = (void *)data };
	ez_blob_dump(&blob, prefix);
}

static void blob_print(const ez_blob_desc *d, const void *a, int depth) {
	ez_blob_print(d, a, depth);
}

static int blob_equals(const ez_blob_desc *desc, const void *a, const void *b) {
	int e = 1;
	for (int i=0, dlen=desc->bd_length; e && i < dlen; i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *u = a + d->bd_offset;
		const void *v = b + d->bd_offset;

		switch (d->bd_type) {
		case EZ_BLOB_INT8:
			e &= (*(uint8_t *)u) == (*(uint8_t *)v);
			break;
		case EZ_BLOB_INT16:
			e &= (*(uint16_t *)u) == (*(uint16_t *)v);
			break;
		case EZ_BLOB_INT32:
			e &= (*(uint32_t *)u) == (*(uint32_t *)v);
			break;
		case EZ_BLOB_INT64:
			e &= (*(uint64_t *)u) == (*(uint64_t *)v);
			break;
		case EZ_BLOB_FLOAT32:
			e &= (*(float *)u) == (*(float *)v);
			break;
		case EZ_BLOB_FLOAT64:
			e &= (*(double *)u) == (*(double *)v);
			break;

		case EZ_BLOB_CSTRING | EZ_BLOB_INT8 | EZ_BLOB_ISNULLABLE:
			if (!*(char **)u || !*(char **)v) {
				e &= (*(char **)u) == (*(char **)v);
				break;
			}
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8:
			if (!*(char **)u || !*(char **)v) {
				printf("strings can't be null\n");
				e = 0;
			} else {
				e &= strcmp(*(char **)u, *(char **)v) == 0;
			}
			break;

		case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64: {
			const struct ez_blob *ua = u;
			const struct ez_blob *va = v;

			e &= ua->eb_size == va->eb_size
				&& memcmp(ua->eb_data, va->eb_data, ua->eb_size) == 0;
			break;
		}

		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP | EZ_BLOB_ISNULLABLE:
			if (!*(void **)u || !*(void **)v) {
				e &= *(void **)u == *(void **)v;
				break;
			}
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
			u = *(void **)u;
			v = *(void **)v;
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
			if (!u || !v) {
				printf("structs p can't be null\n");
				e = 0;
			} else {
				e &= blob_equals(d->bd_table, u, v);
			}
			break;


		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP | EZ_BLOB_ISNULLABLE:
			if (!*(void **)u || !*(void **)v) {
				e &= *(void **)u == *(void **)v;
				break;
			}
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
			u = *(void **)u;
			v = *(void **)v;
		case EZ_BLOB_STRUCT | EZ_BLOB_LIST:
			if (!u || !v) {
				printf("lists p can't be null\n");
				e = 0;
			} else {
				ez_list *ul = (void *)u;
				ez_list *vl = (void *)v;


				e &= ez_list_size(ul) == ez_list_size(vl);
				for (ez_node *uw = ez_list_head(ul), *un = ez_node_succ(uw),
					     *vw = ez_list_head(vl), *vn = ez_node_succ(uw);
				     e && un && vn;
				     uw = un, un = ez_node_succ(un),
					     vw = vn, vn = ez_node_succ(vn)) {
					e &= blob_equals(d->bd_table, uw, vw);
				}
			}
			break;

		default:
			printf("cmp unknown field type=%d id=%d\n", d->bd_type, d->bd_id);
			e = 0;
			break;
		}

		if (!e) {
			printf("cmp failed in struct %d at field %d type %04x\n", desc->bd_id, d->bd_id, d->bd_type);
			abort();
		}
	}

	return e;
}

struct test_funcs {
	const char *name;
	int testno;
	int (*blob_encode)(const ez_blob_desc *d, const void *p, ez_blob *blob);
	void *(*blob_decode)(const ez_blob_desc *desc, const ez_blob *blob);
	int (*blob_decode_raw)(const ez_blob_desc *desc, const ez_blob *blob, void *p);
};

static void test_basics(const ez_blob_desc *d, void *src, struct test_funcs *funcs) {
	ez_blob blob, blob2;
	void *t;
	size_t dsize = d->bd_offset;
	char tmp[dsize + 64] __attribute__ ((aligned(64)));
	int res;

	if (doprint) {
		printf("source\n");
		blob_print(d, src, 4);
	}

	res = funcs->blob_encode(d, src, &blob);
	if (res == ENOMEM) {
		printf(" encode failed: probably out of memory, skipped testing this\n");
		return;
	}
	assert(res == 0);

	if (doprint)
		dumphex(blob.eb_data, blob.eb_size, "serial: ");
	else
		dumphex(blob.eb_data, blob.eb_size < 256 ? blob.eb_size : 256, "header: ");

	// check decode equals
	t = funcs->blob_decode(d, &blob);
	assert(t != NULL);

	if (doprint) {
		printf("decoded\n");
		blob_print(d, t, 4);
	}
	assert(blob_equals(d, t, src));

	// check encode-decoded equals source
	res = funcs->blob_encode(d, t, &blob2);
	assert(res == 0);
	assert(blob.eb_size == blob2.eb_size);
	assert(memcmp(blob.eb_data, blob2.eb_data, blob2.eb_size) == 0);
	free(blob2.eb_data);

	ez_blob_free(d, t);

	// check raw decode stays within bounds of struct
	memset(tmp, 0xbe, sizeof(tmp));
	funcs->blob_decode_raw(d, &blob, &tmp[32]);
	//assert(t != NULL);
	//assert(blob_equals(d, tmp+32, src));
	if (doprint)
		dumphex(tmp, sizeof(tmp), "object: ");

	for (int i=0;i<32;i++) {
		assert((tmp[i] & 0xff) == 0xbe);
		assert((tmp[i + 32 + dsize] & 0xff) == 0xbe);
	}
	ez_blob_free_raw(d, tmp+32);
	free(blob.eb_data);
}

static void test_simple(struct test_funcs *funcs) {
	struct simple src = {
		'z', 0xff0f, 1, 0xf01040a04010dULL, 3.14159f, 5644941221133, "hello test"
	};
	const ez_blob_desc *d = simple_DESC;

	printf("test simple\n");
	test_basics(d, &src, funcs);
}

static void test_arrays(struct test_funcs *funcs) {
	char data[7] = { 1, 2, 3, 4, 5, 6, 7 };
	struct arrays src = {
		32,
		"Bob was here.",
		.array.eb_size = 7,
		.array.eb_data = data
	};
	const ez_blob_desc *d = array_DESC;

	printf("test arrays\n");
	test_basics(d, &src, funcs);
}

static void test_struct(struct test_funcs *funcs) {
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23, "foobar"
		},
		{
			55, "Jane was there.", .array.eb_size = 10, .array.eb_data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;

	printf("test embedded structures\n");
	test_basics(d, &src, funcs);
}

static void test_rawtree(struct test_funcs *funcs) {
	struct tree src[6] = {
		{ &src[1], &src[3], "root" },
		{ NULL, &src[2], "left" },
		{ NULL, NULL, "left-right" },
		{ &src[4], &src[5], "right" },
		{ NULL, NULL, "right-left" },
		{ NULL, NULL, "right-right" },
	};
	const ez_blob_desc *d = tree_DESC;

	printf("test raw tree\n");
	test_basics(d, &src, funcs);
}

static void test_list(struct test_funcs *funcs) {
	struct list list = {
		.name = "a list",
		.list = EZ_INIT_LIST(list.list)
	};
	struct listnode nodes[3] = {
		{ .name = "node 0", .value.eb_size = 5, .value.eb_data = "data0" },
		{ .name = "node 1", .value.eb_size = 5, .value.eb_data = "data1" },
		{ .name = "node 2", .value.eb_size = 5, .value.eb_data = "data2" },
	};

	list.name = "a list";
	//	ez_list_init(&list.
	for (int i=0;i<3;i++)
		ez_list_addtail(&list.list, &nodes[i]);

	const ez_blob_desc *d = list_DESC;

	printf("test list\n");
	test_basics(d, &list, funcs);
}

static void test_intsize(struct test_funcs *funcs) {
	uint64_t sizes[8] = { 0x00, 0xff, 0x101, 0xffff, 0x10002, 0xffffffff, 0x100000003ULL, ~0ULL };
	const ez_blob_desc *d = intsize_DESC;
	struct intsize src;

	// Checks that the sizes reconstruct to their original values

	for (int i=0;i<8;i++) {
		uint64_t size = sizes[i];

		src.value16 = size;
		src.value32 = size;
		src.value64 = size;

		printf("test int value %016lx\n", size);
		test_basics(d, &src, funcs);
	}
}

static void test_stringsize(struct test_funcs *funcs) {
	size_t sizes[7] = { 0, 1, 255, 256, 65535, 65536, 0x100000000ULL };
	const ez_blob_desc *d = stringsize_DESC;
	struct stringsize src;
	int sprint = doprint;

	for (int i=0;i<7;i++) {
		size_t size = sizes[i];

		// only tagz supports >32 bit counts
		if (!dolarge
		    || (size >= 0x100000000ULL
			&& strcmp(funcs->name, "tagz") != 0)) {
			printf("skip string size %zd\n", size);
			continue;
		}

		src.value = malloc(size+1);
		if (src.value) {
			memset(src.value, ' ', size);
			src.value[size] = 0;
			printf("test string size %zd\n", size);
			doprint = size <= 256;
			test_basics(d, &src, funcs);
			free(src.value);
		} else {
			printf("skip string size %zd: memory allocation failed\n", size);
		}
	}
	doprint = sprint;
}

// the corruption tests don't know the format details so can't tell what a given
// change should cause.
// however they should not crash, cause leaks, or valgrind oddness.

static void test_corrupt_bit(struct test_funcs *funcs) {
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23, "hello goodbyte"
		},
		{
			55, "Jane was there.", .array.eb_size = 10, .array.eb_data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;
	struct embedded *dst;
	int res;
	ez_blob blob;

	// test 1 corrupt bit in every position

	// we don't know the internals so which ones would fail, but
	// can check for bad memory accesses with valgrind or crash

	int good = 0, bad = 0;
	res = funcs->blob_encode(d, &src, &blob);
	assert(res == 0);
	for (int i=0;i<blob.eb_size;i++) {
		uint8_t *data = blob.eb_data;
		for (int j=0;j<8;j++) {
			data[i] ^= (1<<j);
			dst = funcs->blob_decode(d, &blob);
			data[i] ^= (1<<j);

			if (dst) {
				good++;
				ez_blob_free(d, dst);
			} else {
				bad++;
			}
		}
	}
	free(blob.eb_data);
	printf(" info: %d failed, %d decoded - but not important\n", bad, good);
}

static void test_corrupt_byte(struct test_funcs *funcs) {
	char data[10] = { 1, 2, 3, 4, 5, 6, 7, 1, 2, 3 };
	struct embedded src = {
		{
			-1, -1, -1, -1, 1337.4004, 6.022E23, "lol"
		},
		{
			55, "Jane was there.", .array.eb_size = 10, .array.eb_data = data
		}
	};
	const ez_blob_desc *d = embed_DESC;
	struct embedded *dst;
	int res;
	ez_blob blob;

	int good = 0, bad = 0;
	res = funcs->blob_encode(d, &src, &blob);
	assert(res == 0);
	for (int i=0;i<blob.eb_size;i++) {
		uint8_t *data = blob.eb_data;
		uint8_t save = data[i];

		for (int j=0;j<256;j++) {
			data[i] = j;
			dst = funcs->blob_decode(d, &blob);
			if (dst) {
				good++;
				ez_blob_free(d, dst);
			} else {
				bad++;
			}
		}
		data[i] = save;
	}
	free(blob.eb_data);
	printf(" info: %d failed, %d decoded - but not important\n", bad, good);
}

struct test_funcs backends[] = {
	{
		"tagz", 0,
		ez_tagz_encode,
		ez_tagz_decode,
		ez_tagz_decode_raw
	},
	{
		"xdrn", 0,
		ez_xdrn_encode,
		ez_xdrn_decode,
		ez_xdrn_decode_raw
	},
	{
		"basic", 1,
		ez_basic_encode,
		ez_basic_decode,
		ez_basic_decode_raw
	},
};

struct {
	const char *name;
	void (*test)(struct test_funcs *);
} tests[] = {
	{ "primitives", test_simple },
	{ "arrays", test_arrays },
	{ "structs", test_struct },
	{ "tree as links", test_rawtree },
	{ "lists", test_list },
	{ "string sizes", test_stringsize },
	{ "int sizes", test_intsize },
	{ "corrupted bit", test_corrupt_bit },
	{ "corrupted byte", test_corrupt_byte },
};

int main(int argc, char **argv) {
	int nfunc = sizeof(backends)/sizeof(backends[0]);
	int ntest = sizeof(tests)/sizeof(tests[0]);

	dolarge = (argc >= 2);

	for (int i=0;i<nfunc;i++) {
		struct test_funcs *f = &backends[i];

		for (int j=0;j<ntest && (f->testno == 0 || j < f->testno);j++) {
			printf("* **********************************************************************\n");
			printf("* test: %s / %s\n", f->name, tests[j].name);
			fflush(stdout);

			tests[j].test(f);
		}
	}
	return 0;
}
